/*   This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
   
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     (C) 2006 schan
*/
#include "SDL/SDL.h"
#include "defines.h"
#include "plaintextrenderer.h"
#include "util.h"
#include <string>
#include <iostream>
#include <fstream>
#include "math.h"

int LineInfo::max_width = 0;
BitmapFont *LineInfo::font = NULL;
bool LineInfo::justify = false;
Uint16 LineInfo::space_char = 0;

LineInfo::LineInfo()
{
	glyphs = NULL;
	num_glyphs = 0;
	formatted = false;
	needs_justify = false;
}

LineInfo::~LineInfo()
{
	delete [] glyphs;
}

void LineInfo::set_max_width(int w)
{
	max_width = w;
}

void LineInfo::set_font(BitmapFont *f)
{
	font = f;
}
		
void LineInfo::set_justify(bool j)
{
	justify = j;
}

void LineInfo::set_space_char(Uint16 c)
{
	space_char = c;
}


int LineInfo::process_text(Uint16 *unicode_text, int begin)
{
	int total_width = 0;
	int w = 0;
	int i = begin;
	int last_word_end = -1;
	while (unicode_text[i]!=0)
	{
		// detect word start
		if (i>0 && unicode_text[i]==space_char && unicode_text[i-1]!=space_char)
		{
			last_word_end = i-1;
		}
	
		w = font->get_width(unicode_text[i]);
		if (total_width+w > max_width)
		{
			if (last_word_end==-1)
			{
				set_text(unicode_text, begin, i, false);
				return i;
			}
			else
			{
				int end = last_word_end+1;
				set_text(unicode_text, begin, end, true);
				return end+1;
			}
		}
		else
		{
			total_width += w;
			
		}			
		i++;
	}
	set_text(unicode_text, begin, i, false);
	return -1;
}

void LineInfo::set_text(Uint16 *unicode_text, int begin, int end, bool line_needs_justify)
{
	// copy text
	num_glyphs = end-begin;
	glyphs = new Glyph[num_glyphs];
	for (int i=0; i<num_glyphs; i++)
	{
		glyphs[i].unicode_char = unicode_text[i+begin];
	}
	
	needs_justify = line_needs_justify;

}


void LineInfo::format()
{
	int x = 0;
	for (int i=0; i<num_glyphs; i++)
	{
		glyphs[i].x=x;
		x+=font->get_width(glyphs[i].unicode_char);
	}
	int total_width = x;
	
	if (justify && needs_justify)
	{
		int space_left = max_width - total_width;
		
		int num_spaces = 0; // count number of spaces AFTER first non-space character
		bool character_found = false;
		for (int i=0; i<num_glyphs; i++)
		{
			if (glyphs[i].unicode_char!=space_char)
			{
				character_found = true;
			}
			else if (character_found)
			{
				num_spaces++;
			}
		}
			
		double extra_space = ((double)space_left)/((double)num_spaces);
		double inc=0;
		character_found = false;
		for (int i=0; i<num_glyphs; i++)
		{
			if (glyphs[i].unicode_char!=space_char)
			{
				character_found=true;
			}
			else if (character_found)
			{
				inc += extra_space;
			}
			glyphs[i].x = lround(double(glyphs[i].x)+inc);
		}
	}
	
	formatted = true;
}

void LineInfo::render(SDL_Surface *dest, int x, int y)
{
	if (!formatted) format();
	for(int i=0; i<num_glyphs; i++)
	{
		font->draw_unicode(glyphs[i].unicode_char, dest, x+glyphs[i].x, y);
	}
}
	
		
///////////////////////////////////////////////////////////////////////////


PlainTextRenderer::PlainTextRenderer(Options *options)
{
	this->options = options;
	reset();
}

PlainTextRenderer::~PlainTextRenderer()
{
	reset();
}

void PlainTextRenderer::reset()
{
	// delete lines
	for (int i=0; i<lines.size(); i++)
	{
		delete lines.at(i);
	}
	lines.clear();
	
	LineInfo::set_max_width(width-options->get_margin());
	LineInfo::set_font(font);
	
	Uint16 *space_text = util::utf8_to_unicode(" ");
	LineInfo::set_space_char(space_text[0]);
	
	LineInfo::set_justify(options->get_justify());
	
	scroll_y = 0;
	max_scroll = -height;
	
}

void PlainTextRenderer::add_line(std::string line, int encoding)
{
	// preprocess
	if (options->get_indent() && line.length()>=2 && line[0]!=' ')
	{
		line.insert(0,std::string(3,' '));
	}
	// TO DO: replace tabs with spaces
	
	Uint16 * unicode_text = NULL;
	switch (encoding)
	{
		case ENC_LATIN1:
			unicode_text = util::latin1_to_unicode(line.c_str(),0,line.length());
			break;
		case ENC_UTF8:
			unicode_text = util::utf8_to_unicode(line.c_str());
			break;
	}

	int begin = 0;
	do
	{
		LineInfo *new_line = new LineInfo();
		begin = new_line->process_text(unicode_text,begin);
		lines.push_back(new_line);
		max_scroll += font->get_height();
	} while (begin!=-1);
	
}

void PlainTextRenderer::render(SDL_Surface *dest)
{
	SDL_FillRect(dest,NULL,SDL_MapRGB(dest->format,
	                                  font->get_bg_color().r,
	                                  font->get_bg_color().g,
	                                  font->get_bg_color().b));
	                                  
	
	int line_h = font->get_height();
	int first = scroll_y / line_h;
	int i = first;
	int y = 0;
	int x = options->get_margin()>>1;
	
	while (y+line_h<=height && i<lines.size())
	{
		lines.at(i)->render(dest,x,y);
		y+=line_h;
		i++;
	}
	
}

void PlainTextRenderer::set_font(BitmapFont *f)
{
	font = f;
}

void PlainTextRenderer::set_viewport_width(int w)
{
	width = w;
}

void PlainTextRenderer::set_viewport_height(int h)
{
	height = h;
	visible_lines = (font==NULL || font->get_height()==0)? 0 : height/font->get_height();
}

void PlainTextRenderer::scroll_by(int n)
{
	scroll_y += n;
	if (scroll_y>max_scroll) scroll_y=max_scroll;
	if (scroll_y<0) scroll_y = 0;
}


/////////////////////////////////////////////////////


void PlainTextRenderer::next_line()
{
	scroll_by(font->get_height());
}


void PlainTextRenderer::prev_line()
{
	scroll_by(-font->get_height());
}

void PlainTextRenderer::next_page()
{
	scroll_by(font->get_height()*visible_lines);
}

void PlainTextRenderer::prev_page()
{
	scroll_by(-font->get_height()*visible_lines);
}

void PlainTextRenderer::go_to_line(unsigned long int n)
{
	scroll_y = 0;
	scroll_by(font->get_height()*n);
}

void PlainTextRenderer::go_to_percent(float p)
{
	scroll_y = (long int)( ((float)max_scroll)*p /100.0 );
}

float PlainTextRenderer::get_percent()
{
	return ((float)scroll_y)/(float)(max_scroll)*100.0;
}

